#include "dllmain.h"

// from $(FASTMAN92_INCLUDE) directory
#include <Assembly/eInstructionSet.h>
#include <Assembly/UsefulMacros.h>
#include <MemoryAddressCalculator/CMemoryAddressCalculator.h>
#include <MemoryPermission/MemoryPermission.h>

/////////
// from 'fastman92 plugin loader'
#include "../../fastman92 limit adjuster/fastman92 plugin loader/ForOtherProjects/Exports.h"

////////
#include <android/log.h>
#include <limits.h>
#include <dlfcn.h>

static void OutputFormattedDebugStringVargs(const char* format, va_list args)
{
	__android_log_vprint(ANDROID_LOG_INFO, "sample plugin", format, args);
}

void OutputFormattedDebugString(const char* format, ...)
{
	va_list args;
	va_start(args, format);
	OutputFormattedDebugStringVargs(format, args);
}

// Set of functions used to write to protected memory
class CPatch
{
public:
	// Writes data to memory
	// For memory, which is marked as read-only
	// The code sections are marked read-only
	// So if you want to change the code, use this function
	// dwAddress - memory address, where to write the data
	// bData - pointer to data
	// iSize - size of data in bytes
	static void WriteDataToMemory(void* dwAddress, const void* bData, int iSize)
	{
		tMemoryPermissionChangeRequest request;
		request.lpAddress = dwAddress;
		request.dwSize = iSize;
		request.flNewProtect = GetNativeNewProtect(F92_MEM_PAGE_EXECUTE_READWRITE);

		SetMemoryPermission(&request);

		memcpy(dwAddress, bData, iSize);

		// Restore old permission
		if (request.bIsOldProtectSet)
		{
			request.flNewProtect = request.lpflOldProtect;
			SetMemoryPermission(&request);
		}
	}

	// Patches memory data
	static void PatchMemoryData(uintptr_t dwAddress, const void* bData, int iSize)
	{
		WriteDataToMemory((void*)dwAddress, bData, iSize);
	}

	// Patches UINT8 value
	static void PatchUINT8(uintptr_t dwAddress, uint8_t to)
	{
		WriteDataToMemory((void*)dwAddress, &to, sizeof(to));
	}

	// Patches UINT16 value
	static void PatchUINT16(uintptr_t dwAddress, uint16_t to)
	{
		WriteDataToMemory((void*)dwAddress, &to, sizeof(to));
	}

	// Patches UINT32 value
	static void PatchUINT32(uintptr_t dwAddress, uint32_t to)
	{
		WriteDataToMemory((void*)dwAddress, &to, sizeof(to));
	}

	// Patches pointer
	void PatchPointer(uintptr_t dwAddress, const void* to)
	{
		WriteDataToMemory((void*)dwAddress, &to, sizeof(to));
	}

	// Redirects the code, see the comments
	static void RedirectCodeEx(
		eInstructionSet sourceInstructionSet,
		uintptr_t dwAddress,
		const void* to)
	{
		// Thumb trampoline may take 8 bytes (if address is aligned to value of 4) or 10 bytes.
		if (sourceInstructionSet == INSTRUCTION_SET_THUMB)
		{
			char code[12];
			unsigned int sizeOfData = 0;

			if (dwAddress % 4 == 0)
			{
				*(uint32_t*)(code + 0) = 0xF000F8DF;	// // LDR.W             R0, [PC, #0]
				*(const void**)(code + 4) = to;	// pointer, where to jump
				sizeOfData = 8;
			}
			else
			{
				*(uint32_t*)(code + 0) = 0xBF00;	// NOP
				*(uint32_t*)(code + 2) = 0xF000F8DF;	// // LDR.W             R0, [PC, #0]

				*(const void**)(code + 6) = to;	// pointer, where to jump
				sizeOfData = 10;
			}

			CPatch::PatchMemoryData(dwAddress, code, sizeOfData);
		}
		// ARM trampoline takes 8 bytes
		else if (sourceInstructionSet == INSTRUCTION_SET_ARM)
		{
			char code[8];

			*(uint32_t*)(code + 0) = 0xE51FF004;	// // LDR             R0, [PC, #0]
			*(const void**)(code + 4) = to;	// pointer, where to jump
			CPatch::PatchMemoryData(dwAddress, code, sizeof(code));
		}
	}

	void RedirectCodeEx(
		eInstructionSet sourceInstructionSet,
		uintptr_t dwAddress,
		uintptr_t to)
	{
		CPatch::RedirectCodeEx(sourceInstructionSet, dwAddress, (const void*)to);
	}

	// Redirects method
	template<typename T>static void RedirectMethodEx(eInstructionSet instructionSet, uintptr_t dwAddress, T method)
	{
		auto _ptr = method;
		void*& ptr = (void*&)_ptr;
		RedirectCodeEx(instructionSet, dwAddress, ptr);
	}
};

/////////////
static TARGET_THUMB NAKED void FunctionDoingNothing()
{
	__asm(
		".thumb\n"
		"BX LR\n"
		);
}

////////////

// ASLR handling
CMemoryAddressCalculator g_mCalc;	// address calculator
JavaVM* g_javaVM;	// Java virtual machine
void* g_applicationLibHandle;

bool isGame_GTASA_2_0 = false;

extern "C"
{
	// refers to type: tOnPluginLoad
	// which is defined in the plugin loader
	bool DLL_EXPORT OnPluginLoad(
		tPluginStartParams* pPluginStartParams
	)
	{
		// Initialise global variables
		g_javaVM = pPluginStartParams->jvm;
		g_applicationLibHandle = pPluginStartParams->applicationLibHandle;
		//////

		// Initialize a calculator of memory addresses
		g_mCalc.Initialize(0, (uintptr_t)pPluginStartParams->applicationLibBaseAddress);

		// Print to debug output
		OutputFormattedDebugString(
			"OnPluginLoad. Current application library base = 0x%llX",
			(uint64_t)g_mCalc.GetCurrentModuleBase()
		);

		// Check application library identifier
		const char* appLibIdentifier = pPluginStartParams->applicationLibIdentifier;
	
		isGame_GTASA_2_0 = TestLibIdentifier(appLibIdentifier, "SA20");

		if (isGame_GTASA_2_0)
			OutputFormattedDebugString("Running app is GTA SA 2.0");
		else
			OutputFormattedDebugString("Unknown application");

		return true;
	}
	
	// This exported function will get executed when application is initialized.
	// The paths to directories are set up in this stage.
	bool DLL_EXPORT OnApplicationInit()
	{
		///////
		// Path where game files are stored.
		const char* rootDirectory = GetStorageRootDirectoryPath();

		// Sample patching
		if (isGame_GTASA_2_0)	// only for GTA SA 2.0
		{
			// See the FLA source code for more examples
			// For example DynamicLimits.cpp from the FLA

			// Redirect CWanterRender::RenderWater to empty function
			// The water rendering will be disabled.
			{
				uintptr_t funcAddress = (uintptr_t)dlsym(g_applicationLibHandle, "_ZN11CWaterLevel11RenderWaterEv");

				CPatch::RedirectCodeEx(
					GET_INSTRUCTION_SET_FROM_ADDRESS(funcAddress),
					GET_CODE_START(funcAddress),
					(void*)&FunctionDoingNothing
				);

				OutputFormattedDebugString("Water rendering disabled.");
			}
		}

		////////

		OutputFormattedDebugString("OnApplicationInit. Root directory = %s", rootDirectory);

		return true;
	}
}