/*
* Copyright (C) fastman92 <fastman92@gmail.com>, website: http://fastman92.com
* Licensed under the MIT License, see LICENSE at top level directory.
*
*/
#include "Loader.h"

#include "Log.h"
#include "ForOtherProjects/Exports.h"

#include <MemoryAddressCalculator/CMemoryAddressCalculator.h>
#include <MultiPlatformSupport/PlatformGeneralisation.h>

#include "../fastman92 limit adjuster/Source files/Core/CPatch.h"
#include "../fastman92 limit adjuster/Source files/Core/UsefulMacros.h"
#include "../fastman92 limit adjuster/Source files/ForOtherProjects/Common/common.h"

#include <string.h>

#include <android/log.h>
#include <dirent.h>
#include <dlfcn.h>
#include <limits.h>
#include <sys\stat.h> 
#include <stdexcept>
#include <unistd.h>

#include <Array/countof.h>
#include <fileIO.h>

char CPluginLoader::ms_packageName[128] = "";

char CPluginLoader::ms_InitialDirectory[PATH_LIMIT];	// EXE directory on Windows, slash on Android
char CPluginLoader::ms_StorageRootDirectory[PATH_LIMIT];	// EXE directory on Windows, STORAGE_ROOT on Android
char CPluginLoader::ms_StorageRootBaseDirectory[PATH_LIMIT];	// Unset on Windows, STORAGE_ROOT_BASE on Android

char CPluginLoader::ms_CacheDirectory[PATH_LIMIT];	// DEX cache directory

std::vector<tOnApplicationInit> CPluginLoader::ms_applicationInitFuncArray;

char CPluginLoader::ms_applicationIdentifier[4] = { 0 };

jobject CPluginLoader::ms_mainActivityDuringLaunch = 0;

#ifdef IS_PLATFORM_ANDROID
// Sets the path to external storage directory
static void GetPackageName(JNIEnv* env, jobject context, char* strPath, size_t bufferSize)
{
	jclass android_content_Context = env->GetObjectClass(context);
	jmethodID midGetPackageName = env->GetMethodID(android_content_Context, "getPackageName", "()Ljava/lang/String;");
	jstring packageName = (jstring)env->CallObjectMethod(context, midGetPackageName);

	if (packageName)
	{
		const char* packageNameStr = env->GetStringUTFChars(packageName, NULL);
		strncpy(strPath, packageNameStr, bufferSize);

		env->ReleaseStringUTFChars(packageName, packageNameStr);
	}
}

// Sets the path to external storage directory
static void GetExternalStorageDirectory(JNIEnv* env, char* strPath, size_t bufferSize)
{
	jclass classEnvironment = env->FindClass("android/os/Environment");
	jmethodID methodIDgetExternalStorageDirectory = env->GetStaticMethodID(classEnvironment, "getExternalStorageDirectory", "()Ljava/io/File;");
	jobject objectFile = env->CallStaticObjectMethod(classEnvironment, methodIDgetExternalStorageDirectory);

	jclass classFile = env->GetObjectClass(objectFile);
	jmethodID methodIDgetAbsolutePath = env->GetMethodID(classFile, "getAbsolutePath", "()Ljava/lang/String;");
	jstring stringPath = (jstring)env->CallObjectMethod(objectFile, methodIDgetAbsolutePath);

	if (stringPath)
	{
		const char* strPathValueStr = env->GetStringUTFChars(stringPath, NULL);

		strncpy(strPath, strPathValueStr, bufferSize);
		env->ReleaseStringUTFChars(stringPath, strPathValueStr);
	}
}

// Sets the path to DEX cache directory
static void GetCacheDirectory(JNIEnv* env, jobject context, char* strPath, size_t bufferSize)
{
	jobject objectFile;
	bool bReadFromF92launcher = false;
	jmethodID midGetCacheDir;

	jclass classF92launcherSettings = env->FindClass("com/fastman92/main_activity_launcher/Settings");

	if (classF92launcherSettings)
	{
		midGetCacheDir = env->GetStaticMethodID(classF92launcherSettings, "getCacheDir", "(Landroid/content/Context;)Ljava/io/File;");

		if (midGetCacheDir)
		{
			objectFile = env->CallStaticObjectMethod(classF92launcherSettings, midGetCacheDir, context);
			bReadFromF92launcher = true;
		}
	}

	if (!bReadFromF92launcher)
	{
		if (env->ExceptionCheck())
			env->ExceptionClear();

		jclass android_content_Context = env->GetObjectClass(context);

		midGetCacheDir = env->GetMethodID(android_content_Context, "getCacheDir", "()Ljava/io/File;");
		objectFile = (jstring)env->CallObjectMethod(context, midGetCacheDir);
	}

	jclass classFile = env->GetObjectClass(objectFile);

	jmethodID methodIDgetAbsolutePath = env->GetMethodID(classFile, "getAbsolutePath", "()Ljava/lang/String;");
	jstring stringPath = (jstring)env->CallObjectMethod(objectFile, methodIDgetAbsolutePath);

	if (stringPath)
	{
		const char* strPathValueStr = env->GetStringUTFChars(stringPath, NULL);

		strncpy(strPath, strPathValueStr, bufferSize);
		env->ReleaseStringUTFChars(stringPath, strPathValueStr);
	}
}

// Sets the path to external files directory
static void GetExternalFilesDirectory(JNIEnv* env, jobject context, char* strPath, size_t bufferSize)
{
	jobject objectFile;
	bool bReadFromF92launcher = false;
	jmethodID methodIDgetExternalFilesDir;

	jclass classF92launcherSettings = env->FindClass("com/fastman92/main_activity_launcher/Settings");

	if (classF92launcherSettings)
	{
		methodIDgetExternalFilesDir = env->GetStaticMethodID(classF92launcherSettings, "getExternalFilesDir", "(Landroid/content/Context;)Ljava/io/File;");
		
		if (methodIDgetExternalFilesDir)
		{
			objectFile = env->CallStaticObjectMethod(classF92launcherSettings, methodIDgetExternalFilesDir, context);
			bReadFromF92launcher = true;
		}
	}

	if (!bReadFromF92launcher)
	{
		if (env->ExceptionCheck())
			env->ExceptionClear();

		jclass android_content_Context = env->GetObjectClass(context);

		methodIDgetExternalFilesDir = env->GetMethodID(android_content_Context, "getExternalFilesDir", "(Ljava/lang/String;)Ljava/io/File;");

		objectFile = (jstring)env->CallObjectMethod(context, methodIDgetExternalFilesDir, nullptr);
	}
		
	jclass classFile = env->GetObjectClass(objectFile);

	jmethodID methodIDgetAbsolutePath = env->GetMethodID(classFile, "getAbsolutePath", "()Ljava/lang/String;");
	jstring stringPath = (jstring)env->CallObjectMethod(objectFile, methodIDgetAbsolutePath);

	if (stringPath)
	{
		const char* strPathValueStr = env->GetStringUTFChars(stringPath, NULL);

		strncpy(strPath, strPathValueStr, bufferSize);
		env->ReleaseStringUTFChars(stringPath, strPathValueStr);
	}
}

uintptr_t pOriginal_GTAActivity_initTouchSense = 0;

class CPluginLoaderHelper
{
public:

	////////////////////////////
	typedef jboolean(*type_NvEventQueueActivity_init_ptr)(JNIEnv* env, jobject GTAactivity, jboolean paramBoolean);
	static type_NvEventQueueActivity_init_ptr original_NvEventQueueActivity_init;

	// from com.nvidia.devtech.NvEventQueueActivity
	static jboolean NvEventQueueActivity_init(JNIEnv* env, jobject GTAactivity, jboolean paramBoolean)
	{
		jboolean bResult = original_NvEventQueueActivity_init(env, GTAactivity, paramBoolean);
		
		// Set the package name
		GetPackageName(env,
			GTAactivity,
			CPluginLoader::ms_packageName,
			sizeof(CPluginLoader::ms_packageName)
		);

		CPluginLoader::DoAppInitializationStage(env, GTAactivity);

		return bResult;
	}

	////////////////////////////
	// typedef void(*type_GTAActivity_initTouchSense_ptr)(JNIEnv* env, jobject GTAactivity, jobject paramContext);
	// static type_GTAActivity_initTouchSense_ptr original_GTAActivity_initTouchSense;
	#ifdef IS_PLATFORM_ANDROID_ARM32
	static TARGET_ARM NAKED void original_GTAActivity_initTouchSense(JNIEnv* env, jobject GTAactivity, jobject paramContext)
	{
		__asm
		(
			".arm\n"
			"STMFD           SP!, {R3-R6,R11,LR}\n"
			"ADD             R11, SP, #0x14\n"

			ASM_JUMP_TO_ADDRESS_STORED_ON_SYMBOL(pOriginal_GTAActivity_initTouchSense)
			);
	}

	// from com.rockstargames.gtalcs.GTAActivity
	static void GTAActivity_initTouchSense(JNIEnv* env, jobject GTAactivity, jobject paramContext)
	{
		original_GTAActivity_initTouchSense(env, GTAactivity, paramContext);

		// Set the package name
		GetPackageName(env,
			paramContext,
			CPluginLoader::ms_packageName,
			sizeof(CPluginLoader::ms_packageName)
		);

		CPluginLoader::DoAppInitializationStage(env, GTAactivity);
	}
	#endif
	////////////////////////////////
};

CPluginLoaderHelper::type_NvEventQueueActivity_init_ptr CPluginLoaderHelper::original_NvEventQueueActivity_init = nullptr;

static void ReplacePointerToNvEventQueueActivity_init(uintptr_t pInit)
{
	CPluginLoaderHelper::original_NvEventQueueActivity_init =
		*(CPluginLoaderHelper::type_NvEventQueueActivity_init_ptr*)pInit;

	CPatch::PatchPointer(pInit, (void*)&CPluginLoaderHelper::NvEventQueueActivity_init);

	OutputFormattedDebugString("Pointer to com.nvidia.devtech.NvEventQueueActivity.init replaced on 0x%llX%.",
		(uint64_t)pInit
	);
}
#endif

// Sets path by function from loader
void CPluginLoader::GetApplicationLibDirectoryPath(char* libPluginPath)
{
	Dl_info info;
	int info_result = dladdr((const void*)&GetApplicationLibDirectoryPath, &info);

	// remove a filename
	int index_of_last_slash = 0;

	for (unsigned int i = 0; info.dli_fname[i]; i++)
	{
		char c = info.dli_fname[i];

		if (c == '/' || c == '\\')
			index_of_last_slash = i;
	}

	memcpy(libPluginPath, info.dli_fname, index_of_last_slash);
	libPluginPath[index_of_last_slash] = 0;
}

// Initializes
bool CPluginLoader::InitialiseLoading(
	JavaVM* javaVm,
	void* reserved,
	const char* applicationIdentifier,
	void* baseAddressForApplicationLibrary
)
{
	this->ms_applicationInitFuncArray.clear();

	CMemoryAddressCalculator memCalc;
	memCalc.Initialize(0, (uintptr_t)baseAddressForApplicationLibrary);

	this->m_javaVm = javaVm;
	this->m_reserved = reserved;
	memcpy(this->ms_applicationIdentifier, applicationIdentifier, sizeof(this->ms_applicationIdentifier));
	this->m_baseAddressForApplicationLibrary = baseAddressForApplicationLibrary;

	if (!CPatch::IsInitialized())
		CPatch::Init();

	// Replace JNI init function
	{
		MAKE_DEAD_IF();
		#ifdef IS_PLATFORM_ANDROID_ARM32
		else if (TestLibIdentifier(ms_applicationIdentifier, "BL18"))	// Bully Scholarship Edition 1.0.0.18
		{
			ReplacePointerToNvEventQueueActivity_init(
				memCalc.GetCurrentVAbyPreferedVA(0xB72FC0)
			);
		}
		else if (TestLibIdentifier(applicationIdentifier, "G_18"))	// GTA III 1.8
		{
			ReplacePointerToNvEventQueueActivity_init(
				memCalc.GetCurrentVAbyPreferedVA(0x351A48)
			);
		}
		else if (TestLibIdentifier(applicationIdentifier, "VC19"))	// GTA VC 1.09
		{
			ReplacePointerToNvEventQueueActivity_init(
				memCalc.GetCurrentVAbyPreferedVA(0x3D20D8)
			);
		}
		else if (TestLibIdentifier(applicationIdentifier, "SA18"))	// GTA SA 1.08
		{
			ReplacePointerToNvEventQueueActivity_init(
				memCalc.GetCurrentVAbyPreferedVA(0x5DDC60)
			);
		}
		else if (TestLibIdentifier(applicationIdentifier, "SA20") // GTA SA 2.00
			|| TestLibIdentifier(applicationIdentifier, "SG29")	// // GTA SA GER 2.09
		)
		{
			ReplacePointerToNvEventQueueActivity_init(
				memCalc.GetCurrentVAbyPreferedVA(0x661E48)
			);
		}
		else if (TestLibIdentifier(applicationIdentifier, "LC24"))
		{
			pOriginal_GTAActivity_initTouchSense = memCalc.GetCurrentVAbyPreferedVA(0x248A50);

			uintptr_t Address_GTAActivity_initTouchSense = memCalc.GetCurrentVAbyPreferedVA(0x248A48);

			CPatch::RedirectCodeEx(
				INSTRUCTION_SET_ARM,
				Address_GTAActivity_initTouchSense,
				(const void*)&CPluginLoaderHelper::GTAActivity_initTouchSense
			);

			OutputFormattedDebugString(
				"Function Java_com_rockstargames_gtalcs_GTAActivity_initTouchSense (0x%llX) redirected.",
				(uint64_t)Address_GTAActivity_initTouchSense
			);
		}
		#endif
		else
		{
			OutputFormattedDebugString("Missing code for initialisation of plugin loader for the current application version.");
			return false;
		}

		return true;
	}
}

// Does plugin loading
void CPluginLoader::DoPluginLoading(
	const char* libPluginPath
)
{
	tPluginStartParams startParams;
	startParams.structureSize = sizeof(tPluginStartParams);
	startParams.jvm = this->m_javaVm;
	startParams.second_param_reserved = this->m_reserved;
	memcpy(startParams.applicationLibIdentifier, this->ms_applicationIdentifier, sizeof(startParams.applicationLibIdentifier));
	startParams.applicationLibBaseAddress = this->m_baseAddressForApplicationLibrary;

	// get library handle of game
	Dl_info applicationModuleInfo;

	if (!dladdr(this->m_baseAddressForApplicationLibrary, &applicationModuleInfo))
	{
		OutputFormattedDebugString("Can't get an information about application library.");
		return;
	}

	const char* applicationModuleFilename = GetFilenameFromPath(applicationModuleInfo.dli_fname);

	void* applicationLibHandle = dlopen(applicationModuleFilename, 4);	// 4 - RTLD_NOLOAD		

	////////////////////////////
	char path[PATH_MAX];
	void *pluginHandle = nullptr;
	void* FLApluginHandle = nullptr;

	bool bCLEOexists = false;
	DIR* dir = opendir(libPluginPath);
	dirent* drnt;

	while ((drnt = readdir(dir)) != nullptr)
	{
		if (!strcmp(drnt->d_name, "libcleo.so"))
		{
			bCLEOexists = true;
			continue;
		}

		int l = strlen(drnt->d_name);

		if (l < 12)
			continue;

		if (memcmp(drnt->d_name, "libplugin", 9)
			|| memcmp(drnt->d_name + l - 3, ".so", 3))
			continue;

		sprintf(path, "%s/%s", libPluginPath, drnt->d_name);

		OutputFormattedDebugString("Loading %s", path);

		pluginHandle = dlopen(path, RTLD_GLOBAL | RTLD_LAZY);

		if (!strcmp(drnt->d_name, FLA_FILENAME))
		{
			OutputFormattedDebugString("The FLA was just loaded.");
			FLApluginHandle = pluginHandle;
		}

		OutputFormattedDebugString("Handle of shared object: 0x%llX", (uint64_t)pluginHandle);

		// OnPluginLoad
		tOnPluginLoad pOnPluginLoad = (tOnPluginLoad)dlsym(pluginHandle, "OnPluginLoad");		

		if (pOnPluginLoad)
		{
			OutputFormattedDebugString("Calling OnPluginLoad from %s", drnt->d_name);

			startParams.pluginHandle = pluginHandle;
			startParams.pluginLibraryPath = path;
			startParams.applicationLibHandle = applicationLibHandle;

			pOnPluginLoad(
				&startParams
			);
		}
		else
			OutputFormattedDebugString("Can't find OnPluginLoad in %s", drnt->d_name);

		// OnApplicationInit
		tOnApplicationInit pOnApplicationInit = (tOnApplicationInit)dlsym(pluginHandle, "OnApplicationInit");;

		if (pOnApplicationInit)
		{
			this->ms_applicationInitFuncArray.push_back(pOnApplicationInit);
			OutputFormattedDebugString("Found OnApplicationInit in %s", drnt->d_name);
		}
		else
			OutputFormattedDebugString("Can't find OnApplicationInit in %s", drnt->d_name);
	}

	closedir(dir);

	// Load CLEO library if it exists
	if (bCLEOexists)
	{
		bool bLoadCLEO = true;

		// delay the responsibility of loading the CLEO to the FLA
		if (FLApluginHandle)
		{
			// Returns true if CLEO loading is implemented by the FLA
			bool (F92_LA_CONVENTION* IsCLEOloadingImplementedByTheFLA)()
				= (bool(F92_LA_CONVENTION*)())
				dlsym(FLApluginHandle, "IsCLEOloadingImplementedByTheFLA");

			if (IsCLEOloadingImplementedByTheFLA)
				bLoadCLEO = !IsCLEOloadingImplementedByTheFLA();
		}

		if (bLoadCLEO)
		{
			sprintf(path, "%s/libcleo.so", libPluginPath);
			OutputFormattedDebugString("Loading CLEO: %s", path);

			pluginHandle = dlopen(path, RTLD_GLOBAL | RTLD_LAZY);

			type_JNI_OnLoad pJni_OnLoad = (type_JNI_OnLoad)dlsym(pluginHandle, "JNI_OnLoad");

			if (pJni_OnLoad)
				pJni_OnLoad(this->m_javaVm, this->m_reserved);
		}
		else
			OutputFormattedDebugString("CLEO should not be loaded by plugin loader.");
	}
	else
		OutputFormattedDebugString("CLEO does not exist.");
}

// Does application initialization stage
void CPluginLoader::DoAppInitializationStage(JNIEnv* jniEnv, jobject mainActivity)
{
	ms_mainActivityDuringLaunch = mainActivity;

	GetCacheDirectory(jniEnv,
		mainActivity,
		CPluginLoader::ms_CacheDirectory,
		sizeof(CPluginLoader::ms_CacheDirectory)
	);

	/////////////
	
	try
	{
		if (TestLibIdentifier(ms_applicationIdentifier, "BL18")	// Bully Scholarship Edition 1.0.0.18
			|| TestLibIdentifier(ms_applicationIdentifier, "G_18")	// GTA III 1.8
			|| TestLibIdentifier(ms_applicationIdentifier, "VC19")	// GTA VC 1.09
			|| TestLibIdentifier(ms_applicationIdentifier, "SA18") 	// GTA SA 1.08
			|| TestLibIdentifier(ms_applicationIdentifier, "SA20")	// GTA SA 2.00
			|| TestLibIdentifier(ms_applicationIdentifier, "SG29")	// GTA SA GER 2.09
			)
		{
			jclass NvUtilClass = jniEnv->FindClass("com/nvidia/devtech/NvUtil");

			if (!NvUtilClass)
				throw std::runtime_error("Cannot find com.nvidia.devtech.NvUtil class.");

			jmethodID g_hasAppLocalValue = jniEnv->GetMethodID(NvUtilClass, "hasAppLocalValue", "(Ljava/lang/String;)Z");
			jmethodID g_getAppLocalValue = jniEnv->GetMethodID(NvUtilClass, "getAppLocalValue", "(Ljava/lang/String;)Ljava/lang/String;");
			jmethodID g_setAppLocalValue = jniEnv->GetMethodID(NvUtilClass, "setAppLocalValue", "(Ljava/lang/String;Ljava/lang/String;)V");
			jmethodID g_getParameter = jniEnv->GetMethodID(NvUtilClass, "getParameter", "(Ljava/lang/String;)Ljava/lang/String;");

			jmethodID getInstance = jniEnv->GetStaticMethodID(NvUtilClass, "getInstance", "()Lcom/nvidia/devtech/NvUtil;");

			if (!getInstance)
				throw std::runtime_error("Cannot find com.nvidia.devtech.NvUtil.getInstance static method.");

			jobject g_globalThiz = jniEnv->CallStaticObjectMethod(NvUtilClass, getInstance);

			if (!g_globalThiz)
				throw std::runtime_error("Function com.nvidia.devtech.NvUtil.getInstance() returned null.");

			/////////////////
			jstring keyName;
			jstring keyValue;

			// STORAGE_ROOT
			keyName = jniEnv->NewStringUTF("STORAGE_ROOT");
			keyValue = (jstring)jniEnv->CallObjectMethod(
				g_globalThiz,
				g_getAppLocalValue,
				keyName
			);

			jniEnv->DeleteLocalRef(keyName);

			if (keyValue)
			{
				const char* keyValueStr = jniEnv->GetStringUTFChars(keyValue, NULL);
				strncpy(CPluginLoader::ms_StorageRootDirectory, keyValueStr, _countof(CPluginLoader::ms_StorageRootDirectory));
				jniEnv->ReleaseStringUTFChars(keyValue, keyValueStr);
			}

			// STORAGE_ROOT_BASE
			keyName = jniEnv->NewStringUTF("STORAGE_ROOT_BASE");
			keyValue = (jstring)jniEnv->CallObjectMethod(
				g_globalThiz,
				g_getAppLocalValue,
				keyName
			);

			jniEnv->DeleteLocalRef(keyName);

			if (keyValue)
			{
				const char* keyValueStr = jniEnv->GetStringUTFChars(keyValue, NULL);
				strncpy(CPluginLoader::ms_StorageRootBaseDirectory, keyValueStr, _countof(CPluginLoader::ms_StorageRootBaseDirectory));
				jniEnv->ReleaseStringUTFChars(keyValue, keyValueStr);
			}
		}
		else if (TestLibIdentifier(ms_applicationIdentifier, "LC24"))
		{
			/*
			char* g_AppSupportPath = (char*)dlsym(
				this->hModule_of_game.linux_os,
				"g_AppSupportPath"
			);

			char* g_WADFilePath = (char*)dlsym(
				this->hModule_of_game.linux_os,
				"g_WADFilePath"
			);
			*/

			GetExternalStorageDirectory(jniEnv, CPluginLoader::ms_StorageRootBaseDirectory, sizeof(CPluginLoader::ms_StorageRootBaseDirectory));
			
			GetExternalFilesDirectory(jniEnv, mainActivity, CPluginLoader::ms_StorageRootDirectory, sizeof(CPluginLoader::ms_StorageRootDirectory));

			/*
			strncpy(CPluginLoader::ms_StorageRootDirectory, CPluginLoader::ms_StorageRootBaseDirectory, sizeof(CPluginLoader::ms_StorageRootDirectory));

			char* pStr = CPluginLoader::ms_StorageRootDirectory + strlen(CPluginLoader::ms_StorageRootDirectory);

			// FASTMAN92_UNSAFE_CODE
			// possible buffer overflow.
			strcpy(pStr, "/Android/data/");
			pStr += strlen(pStr);

			strcpy(pStr, CPluginLoader::ms_packageName);
			pStr += strlen(pStr);

			strcpy(pStr, "/files");
			pStr += strlen(pStr);
			*/
		}
		else
			throw std::runtime_error("Missing code to set paths for current game version.");

		getcwd(CPluginLoader::ms_InitialDirectory, sizeof(CPluginLoader::ms_InitialDirectory));

		////
		OutputFormattedDebugString(
			"Root base directory: %s Root directory: %s Initial directory: %s",
			CPluginLoader::ms_StorageRootBaseDirectory,
			CPluginLoader::ms_StorageRootDirectory,
			CPluginLoader::ms_InitialDirectory
		);

		// Create root directory if it does not exist
		/*
		struct stat st;
		if (stat(CPluginLoader::ms_StorageRootDirectory, &st) != 0 || !(st.st_mode & S_IFDIR))
		{
			int mkdir_result = mkdir_p(CPluginLoader::ms_StorageRootDirectory, 0771);
			
			if (mkdir_result != 0)
				throw std::runtime_error("An error while creating game root directory.");
			else
			{
				OutputFormattedDebugString("Created directory");
			}
		}
		else
		{
			OutputFormattedDebugString("already exists");
		}
		*/

		// Set current working directory
		if (!SetCurrentWorkingDirectory_OS_independent(CPluginLoader::ms_StorageRootDirectory))
			throw std::runtime_error("An error while setting the current working directory to game root directory.");
	}
	catch (const std::runtime_error& e)
	{
		OutputFormattedDebugString("%s", e.what());
		return;
	}

	///////////////////////////////
	
	// Execute OnApplicationInit functions
	for (tOnApplicationInit func : CPluginLoader::ms_applicationInitFuncArray)
		func();
}